#!/usr/bin/python
# -*- coding: utf-8 -*-

from postgres_ddl.Config import Config
from postgres_ddl.DDL import DDL

class Extension(DDL):
    def __init__(self, parent, data):
        super().__init__(parent, data)

        self.Oid = data.get("oid")
        assert (self.Oid or 0) > 0, \
            "Extension oid is null"

        self.Schema = (data.get("schema") or "").strip()
        assert len(self.Schema) > 0, \
            "Extension schema is null"

        self.Name = (data.get("name") or "").strip()
        assert len(self.Name) > 0, \
            "Extension name is null"

        self.Version = (data.get("version") or "").strip()
        assert len(self.Version) > 0, \
            "Extension version is null"

        self.Owner = (data.get("owner") or "").strip()
        assert len(self.Owner) > 0, \
            "Extension owner is null"

    def __str__(self):
        return self.GetFullName()

    def GetObjectType(self):
        return "extension"

    def GetObjectName(self):
        return f"{self.GetObjectType()}_{self.GetFullName()}"

    def GetFullName(self):
        return f"{self.Name}"

    def GetTag(self):
        return "EXTENSION"

    def DDL_Drop(self):
        return f"DROP {self.GetTag()} IF EXISTS {self.GetFullName()};"

    def DDL_Create(self):
        r = f"-- Extension: {self.GetFullName()}"
        r += Config.NL + Config.NL
        r += f"-- {self.DDL_Drop()}"
        r += Config.NL + Config.NL
        r += f"SET ROLE {self.Owner};"
        r += Config.NL + Config.NL
        r += f"CREATE {self.GetTag()} {self.GetFullName()}"
        r += Config.NL
        r += f'{Config.Indent}SCHEMA "{self.Schema}"'
        r += Config.NL
        r += f'{Config.Indent}VERSION "{self.Version}"'
        r += Config.NL
        r = r.strip() + ";"
        return r.strip() + Config.NL

    def GetPath(self):
        return ["_extension"]

    def GetFileName(self):
        return f"{self.Name}.sql"

    def Export(self):
        return { self.GetObjectName() : self }

    def Diff(self, another):
        if self.DDL_Create() != another.DDL_Create():
            return [
                another.DDL_Drop(),
                self.DDL_Create()
            ]

        return []



#!/usr/bin/python
# -*- coding: utf-8 -*-

import logging

import psycopg2
from psycopg2.extensions import ISOLATION_LEVEL_READ_COMMITTED
from psycopg2.extras import RealDictCursor

from postgres_ddl.Cast import Cast
from postgres_ddl.Config import Config
from postgres_ddl.EventTrigger import EventTrigger
from postgres_ddl.Extension import Extension
from postgres_ddl.ForeignServer import ForeignServer
from postgres_ddl.ForeignTable import ForeignTable
from postgres_ddl.Function import Function
from postgres_ddl.Namespace import Namespace
from postgres_ddl.Procedure import Procedure
from postgres_ddl.Publication import Publication
from postgres_ddl.Sequence import Sequence
from postgres_ddl.Subscription import Subscription
from postgres_ddl.System import CalcDuration
from postgres_ddl.Table import Table
from postgres_ddl.View import View

class Database():
    def __init__(self, connect, exclude_schemas=[]):
        """
            Database object
            @param connect: Connection params
            @param exclude_schemas: Excluded schemas (namespaces)
        """
        self.Connect        = self.GetConnectionString(connect)
        self.IsSuperUser    = False
        self.Version        = None
        self.Objects        = {}
        self.ExcludeSchemas = exclude_schemas

    def __str__(self):
        """
            String representation
        """
        return str(self.PG)

    def Parse(self):
        """
            Database metadata (DDL) parsing
        """
        logging.debug("Parsing started = %s", self.Connect)

        with psycopg2.connect(self.Connect) as conn:
            conn.set_isolation_level(ISOLATION_LEVEL_READ_COMMITTED)
            with conn.cursor(cursor_factory = RealDictCursor) as cursor:
                self.GetSuperUser(cursor)
                self.GetVersion(cursor)
                self.GetNamespace(cursor)
                self.GetTable(cursor)
                self.GetFunction(cursor)
                self.GetProcedure(cursor)
                self.GetView(cursor)
                self.GetSequence(cursor)
                self.GetForeignServer(cursor)
                self.GetForeignTable(cursor)
                self.GetExtension(cursor)
                self.GetPublication(cursor)
                self.GetSubscription(cursor)
                self.GetEventTrigger(cursor)
                self.GetCast(cursor)

    def GetConnectionString(self, connect):
        assert connect is not None, \
            "connection params is null"

        assert isinstance(connect, dict), \
            "connection params is not dict"

        assert len(connect.keys()) > 0, \
            "connection dict is empty"

        host = (connect.get("host") or "").strip()
        assert len(host) > 0, \
            "host name is empty"

        port = connect.get("port") or 5432

        database = (connect.get("database") or "").strip()
        assert len(database) > 0, \
            "database name is empty"

        username = (connect.get("username") or "").strip()
        assert len(username) > 0, \
            "username is empty"

        password = (connect.get("password") or "").strip()
        assert len(password) > 0, \
            "password is empty"

        return "host={0} port={1} dbname={2} user={3} password={4}".format(
            host, port, database, username, password)

    @CalcDuration
    def GetSuperUser(self, cursor):
        cursor.execute("""select exists(select 1 from pg_roles where rolname = user and rolsuper) as is_super""")

        for row in cursor.fetchall():
            self.IsSuperUser = row.get("is_super") or False

        logging.info("IsSuperUser = {0}".format(self.IsSuperUser))

    @CalcDuration
    def GetVersion(self, cursor):
        cursor.execute("""
            select trim(replace(v.v, 'PostgreSQL ', ''))::integer as version
            from unnest((select regexp_matches(version(), 'PostgreSQL \\d{1,}'))) v
        """)

        for row in cursor.fetchall():
            self.Version = row.get("version")

        assert (self.Version or 0) > 0, \
            "failed to get PostgreSQL version "

        logging.info("PostgreSQL version = {0}".format(self.Version))

    @CalcDuration
    def GetNamespace(self, cursor):
        cursor.execute("""
            SELECT
                n.oid,
                trim(lower(n.nspname)) AS name,
                trim(lower(r.rolname)) AS owner,
                trim(coalesce(obj_description(n.oid, 'pg_namespace'), '')) AS comment,
                n.nspacl::varchar[] as acl
            FROM pg_namespace n
            JOIN pg_roles r ON
                r.oid = n.nspowner
            WHERE
                n.nspname != ALL(%s) AND
                n.nspname !~* '^pg_temp' AND
                n.nspname !~* '^pg_toast'
            ORDER BY 2,3
        """, [self.ExcludeSchemas])

        for row in cursor.fetchall():
            self.Objects.update(Namespace(None, row).Export())

        logging.info("GetNamespace loaded = {0}".format(cursor.rowcount))

    @CalcDuration
    def GetTableColumn(self, cursor):
        result = {}
        cursor.execute("""
            select
                a.attrelid as table_oid,
                trim(lower(a.attname)) as name,
                trim(lower(format_type(a.atttypid, a.atttypmod))) as type,
                a.attnotnull as not_null,
                pg_get_expr(ad.adbin, ad.adrelid) as default_value,
                trim(d.description) as comment,
                a.attnum as order_num,
                max(a.attnum) over (partition by a.attrelid) as max_order_num
            from pg_attribute a
            left join pg_attrdef ad on
                a.atthasdef and
                ad.adrelid = a.attrelid and
                ad.adnum = a.attnum
            left join pg_description d on
                d.objoid = a.attrelid and
                d.objsubid = a.attnum
            where
                a.attnum > 0 and
                not a.attisdropped
            order by table_oid, order_num
        """, [self.ExcludeSchemas])
        for row in cursor.fetchall():
            table_oid = row.get("table_oid") or 0
            if table_oid not in result.keys():
                result[table_oid] = []
            result.get(table_oid).append(row)

        logging.info("GetTableColumn loaded = {0}".format(cursor.rowcount))

        return result

    @CalcDuration
    def GetTableConstraint(self, cursor):
        result = {}
        cursor.execute("""
            select
                co.conrelid as table_oid,
                trim(lower(co.conname)) as name,
                trim(lower(co.contype)) as type,
                case trim(lower(co.contype))
                    when 'p' then 1
                    when 'u' then 2
                    when 'c' then 3
                    when 'f' then 4
                    else          5
                end::integer as order_num,
                pg_get_constraintdef(co.oid) as definition,
                case trim(lower(co.confupdtype))
                    when 'a' then 'ON UPDATE NO ACTION'
                    when 'r' then 'ON UPDATE RESTRICT'
                    when 'c' then 'ON UPDATE CASCADE'
                    when 'n' then 'ON UPDATE SET NULL'
                    when 'd' then 'ON UPDATE SET DEFAULT'
                end as update_action,
                case trim(lower(co.confdeltype))
                    when 'a' then 'ON DELETE NO ACTION'
                    when 'r' then 'ON DELETE RESTRICT'
                    when 'c' then 'ON DELETE CASCADE'
                    when 'n' then 'ON DELETE SET NULL'
                    when 'd' then 'ON DELETE SET DEFAULT'
                end as delete_action,
                case trim(lower(co.confmatchtype))
                    when 'f' then 'MATCH FULL'
                    when 'p' then 'MATCH PARTIAL'
                    when 'u' then 'MATCH SIMPLE'
                    when 's' then 'MATCH SIMPLE'
                end as match_action,
                case
                    when trim(lower(co.contype)) != 'f' then ''
                    when co.condeferrable and co.condeferred then
                        'DEFERRABLE INITIALLY DEFERRED'
                    when co.condeferrable and not co.condeferred then
                        'DEFERRABLE INITIALLY IMMEDIATE'
                    else
                        'NOT DEFERRABLE'
                end as deferrable_type
            from pg_constraint co
            where co.conislocal
            order by table_oid, name
        """, [self.ExcludeSchemas])
        for row in cursor.fetchall():
            table_oid = row.get("table_oid") or 0
            if table_oid not in result.keys():
                result[table_oid] = []
            result.get(table_oid).append(row)

        logging.info("GetTableConstraint loaded = {0}".format(cursor.rowcount))

        return result

    @CalcDuration
    def GetTableIndex(self, cursor):
        result = {}
        cursor.execute("""
            select
                i.indrelid as table_oid,
                trim(lower(ic.relname)) as name,
                pg_get_indexdef(i.indexrelid, 0, true) as definition
            from pg_index i
            join pg_class ic on
                ic.oid = i.indexrelid
            where
                not i.indisprimary and
                not exists(
                    select 1
                    from pg_constraint co
                    where co.conindid = ic.oid
                )
            order by table_oid, name
        """, [self.ExcludeSchemas])
        for row in cursor.fetchall():
            table_oid = row.get("table_oid") or 0
            if table_oid not in result.keys():
                result[table_oid] = []
            result.get(table_oid).append(row)

        logging.info("GetTableIndex loaded = {0}".format(cursor.rowcount))

        return result

    @CalcDuration
    def GetTableTrigger(self, cursor):
        result = {}
        cursor.execute("""
            select
                tr.tgrelid as table_oid,
                trim(lower(tr.tgname)) as name,
                tr.tgenabled = 'D' as is_disabled,
                pg_get_triggerdef(tr.oid) as definition
            from pg_trigger tr
            where not tr.tgisinternal
            order by table_oid, name
        """, [self.ExcludeSchemas])
        for row in cursor.fetchall():
            table_oid = row.get("table_oid") or 0
            if table_oid not in result.keys():
                result[table_oid] = []
            result.get(table_oid).append(row)

        logging.info("GetTableTrigger loaded = {0}".format(cursor.rowcount))

        return result

    @CalcDuration
    def GetTablePolicy(self, cursor):
        result = {}
        cursor.execute("""
            select
                plc.polrelid as table_oid,
                plc.polname as name,
                case trim(lower(plc.polcmd))
                    when 'r' then 'SELECT'
                    when 'a' then 'INSERT'
                    when 'w' then 'UPDATE'
                    when 'd' then 'DELETE'
                    else 'ALL'
                end as command,
                case when plc.polpermissive then 'PERMISSIVE' else 'RESTRICTIVE' end as type,
                pg_get_expr(plc.polqual, plc.polrelid) as expr_using,
                pg_get_expr(plc.polwithcheck, plc.polrelid) as expr_with_check,
                (
                    select array_agg(distinct trim(lower(coalesce(plcrr.rolname, 'public'))))
                    from unnest(plc.polroles) plcr
                    left join pg_roles plcrr on
                        plcrr.oid = plcr
                ) as roles
            from pg_policy plc
            where
                plc.polqual is not null or
                plc.polwithcheck is not null
            order by table_oid, name
        """, [self.ExcludeSchemas])
        for row in cursor.fetchall():
            table_oid = row.get("table_oid") or 0
            if table_oid not in result.keys():
                result[table_oid] = []
            result.get(table_oid).append(row)

        logging.info("GetTablePolicy loaded = {0}".format(cursor.rowcount))

        return result

    @CalcDuration
    def GetTableSequence(self, cursor):
        result = {}
        cursor.execute("""
            select
                d.refobjid as table_oid,
                n.nspname as schema,
                s.relname as name,
                s.relacl as acl,
                o.rolname as owner,
                trim(coalesce(obj_description(s.oid, 'pg_class'), '')) as comment
            from pg_depend d
            join pg_class s on
                s.oid = d.objid and
                s.relkind = 'S'
            join pg_namespace n on
                n.oid = s.relnamespace AND
                n.nspname != ALL(%s)
            join pg_roles o ON
                o.oid = s.relowner
        """, [self.ExcludeSchemas])
        for row in cursor.fetchall():
            result[row.get("table_oid")] = row

        logging.info("GetTableSequence loaded = {0}".format(cursor.rowcount))

        return result

    @CalcDuration
    def GetTable(self, cursor):
        if self.Version in (9, 9):
            query = """
                SELECT
                    c.oid,
                    trim(lower(n.nspname)) AS schema,
                    trim(lower(c.relname)) AS name,
                    trim(lower(r.rolname)) AS owner,
                    trim(upper(c.relhasoids::varchar)) as has_oids,
                    obj_description(c.oid, 'pg_class') AS comment,
                    case
                        when coalesce(trim(pc.relname), '') = '' then null
                        else pn.nspname || '.' || pc.relname
                    end AS parent_table,
                    null::varchar as part_border,
                    null::varchar as part_key,
                    c.relacl AS acl,
                    c.reloptions
                FROM pg_class c
                JOIN pg_namespace n ON
                    n.oid = c.relnamespace AND
                    n.nspname !~* '^pg_temp' AND
                    n.nspname !~* '^pg_toast'
                JOIN pg_roles r ON
                    r.oid = c.relowner
                LEFT JOIN pg_inherits inh ON
                    c.oid = inh.inhrelid
                LEFT JOIN pg_class pc ON
                    pc.oid = inh.inhparent
                LEFT JOIN pg_namespace pn ON
                    pn.oid = pc.relnamespace
                WHERE
                    c.relkind in ('r','p') AND
                    n.nspname != ALL(%s)
                ORDER BY 2,3
            """
        elif self.Version in (10,11):
            query = """
                SELECT
                    c.oid,
                    trim(lower(n.nspname)) AS schema,
                    trim(lower(c.relname)) AS name,
                    trim(lower(r.rolname)) AS owner,
                    trim(upper(c.relhasoids::varchar)) as has_oids,
                    obj_description(c.oid, 'pg_class') AS comment,
                    case
                        when coalesce(trim(pc.relname), '') = '' then null
                        else pn.nspname || '.' || pc.relname
                    end AS parent_table,
                    pg_get_expr(c.relpartbound, c.oid, true) as part_border,
                    pg_get_partkeydef(c.oid) as part_key,
                    c.relacl::varchar[] AS acl,
                    c.reloptions
                FROM pg_class c
                JOIN pg_namespace n ON
                    n.oid = c.relnamespace AND
                    n.nspname !~* '^pg_temp' AND
                    n.nspname !~* '^pg_toast'
                JOIN pg_roles r ON
                    r.oid = c.relowner
                LEFT JOIN pg_inherits inh ON
                    c.oid = inh.inhrelid
                LEFT JOIN pg_class pc ON
                    pc.oid = inh.inhparent
                LEFT JOIN pg_namespace pn ON
                    pn.oid = pc.relnamespace
                WHERE
                    c.relkind in ('r','p') AND
                    n.nspname != ALL(%s)
                ORDER BY 2,3
            """
        elif self.Version in (12,13,14,15,16):
            query = """
                SELECT
                    c.oid,
                    trim(lower(n.nspname)) AS schema,
                    trim(lower(c.relname)) AS name,
                    trim(lower(r.rolname)) AS owner,
                    false::varchar as has_oids,
                    obj_description(c.oid, 'pg_class') AS comment,
                    case
                        when coalesce(trim(pc.relname), '') = '' then null
                        else pn.nspname || '.' || pc.relname
                    end AS parent_table,
                    pg_get_expr(c.relpartbound, c.oid, true) as part_border,
                    pg_get_partkeydef(c.oid) as part_key,
                    c.relacl::varchar[] AS acl,
                    c.reloptions
                FROM pg_class c
                JOIN pg_namespace n ON
                    n.oid = c.relnamespace AND
                    n.nspname !~* '^pg_temp' AND
                    n.nspname !~* '^pg_toast'
                JOIN pg_roles r ON
                    r.oid = c.relowner
                LEFT JOIN pg_inherits inh ON
                    c.oid = inh.inhrelid
                LEFT JOIN pg_class pc ON
                    pc.oid = inh.inhparent
                LEFT JOIN pg_namespace pn ON
                    pn.oid = pc.relnamespace
                WHERE
                    c.relkind in ('r','p') AND
                    n.nspname != ALL(%s)
                ORDER BY 2,3
            """
        else:
            raise Exception("Unknown PostgreSQL version - {0}".format(self.Version))

        columns = self.GetTableColumn(cursor)
        constraints = self.GetTableConstraint(cursor)
        indexes = self.GetTableIndex(cursor)
        sequences = self.GetTableSequence(cursor)
        triggers = self.GetTableTrigger(cursor)
        policies = self.GetTablePolicy(cursor)

        cursor.execute(query, [self.ExcludeSchemas])
        for row in cursor.fetchall():
            table_oid = row.get("oid") or 0
            row["columns"] = columns.get(table_oid)
            row["constraints"] = constraints.get(table_oid)
            row["sequence"] = sequences.get(table_oid)
            row["indexes"] = indexes.get(table_oid)
            row["triggers"] = triggers.get(table_oid)
            row["policies"] = policies.get(table_oid)
            self.Objects.update(Table(None, row).Export())

        logging.info("GetTable loaded = {0}".format(cursor.rowcount))

    @CalcDuration
    def GetFunction(self, cursor):
        cursor.execute("""
            select
                p.oid,
                trim(lower(n.nspname)) as schema,
                trim(lower(p.proname)) as proc,
                oidvectortypes(proargtypes) as args_in_types,
                pg_get_function_arguments(p.oid) as args_in,
                regexp_replace(regexp_replace(regexp_replace(pg_get_function_result(p.oid), '(?<=[A-Za-z\]])\, ', e'\,\n    ', 'igm'), 'table\(', e'TABLE\(\n    ', 'igm'), '\)$', e'\n\)', 'igm') as args_out,
                coalesce(p.procost, 0) as cost,
                coalesce(p.prorows, 0) as rows,
                trim(lower(o.rolname)) as owner,
                trim(lower(l.lanname)) as lang,
                obj_description(p.oid, 'pg_proc') as comment,
                case
                    when p.provolatile = 'i' then 'IMMUTABLE'
                    when p.provolatile = 's' then 'STABLE'
                    when p.provolatile = 'v' then 'VOLATILE'
                end || case
                    when p.proisstrict then ' STRICT'
                    else ''
                end || case
                    when not p.prosecdef then ''
                    else ' SECURITY DEFINER'
                end as volatility,
                1 < count(*) over (partition by n.nspname, p.proname) as has_duplicate,
                coalesce(trim(lower(t.typname)), '') = 'trigger' as is_trigger,
                replace(p.prosrc, E'\r', '') as code,
                p.proacl::varchar[] as acl
            from pg_proc p
            join pg_namespace n on
                n.oid = p.pronamespace and
                n.nspname !~* '^pg_temp' AND
                n.nspname !~* '^pg_toast' AND
                n.nspname != ALL(%(exclude_schemas)s)
            join pg_language l on
                l.oid = p.prolang and
                l.lanname in ('sql','plpgsql','plpythonu','plpython3u','plproxy')
            join pg_roles o on
                o.oid = p.proowner
            join pg_type t on
                t.oid = p.prorettype
            join lateral(
                select
                    %(indent)s as indent
            ) vrb on true
            where p.prokind in ('f')
            order by 1
        """, {
            "indent": Config.Indent,
            "exclude_schemas": self.ExcludeSchemas
        })

        for row in cursor.fetchall():
            self.Objects.update(Function(None, row).Export())

        logging.info("GetFunction loaded = {0}".format(cursor.rowcount))

    @CalcDuration
    def GetProcedure(self, cursor):
        cursor.execute("""
            select
                p.oid,
                trim(lower(n.nspname)) as schema,
                trim(lower(p.proname)) as proc,
                oidvectortypes(p.proargtypes) as args_in_types,
                pg_get_function_arguments(p.oid) as args_in,
                trim(lower(o.rolname)) as owner,
                trim(lower(l.lanname)) as lang,
                obj_description(p.oid, 'pg_proc') as comment,
                1 < count(*) over (partition by n.nspname, p.proname) as has_duplicate,
                replace(p.prosrc, E'\r', '') as code,
                p.proacl::varchar[] as acl
            from pg_proc p
            join pg_namespace n on
                n.oid = p.pronamespace and
                n.nspname !~* '^pg_temp' AND
                n.nspname !~* '^pg_toast' AND
                n.nspname != ALL(%s)
            join pg_language l on
                l.oid = p.prolang and
                l.lanname in ('sql','plpgsql','plpythonu','plpython3u','plproxy')
            join pg_roles o on
                o.oid = p.proowner
            join pg_type t on
                t.oid = p.prorettype
            where p.prokind in ('p')
            order by 1
        """, [self.ExcludeSchemas])

        for row in cursor.fetchall():
            self.Objects.update(Procedure(None, row).Export())

        logging.info("GetProcedure loaded = {0}".format(cursor.rowcount))

    @CalcDuration
    def GetView(self, cursor):
        query = """
            SELECT
                c.oid,
                trim(lower(n.nspname)) AS schema,
                trim(lower(c.relname)) AS name,
                trim(lower(r.rolname)) AS owner_name,
                trim(coalesce(obj_description(c.oid, 'pg_class'), '')) AS comment,
                pg_get_viewdef(c.oid, true) as definition,
                c.relacl::varchar[] AS acl,
                c.relkind = 'm' as is_materialized,
                &indexes&,
                (
                    select jsonb_agg(q) from (
                        select
                            trim(lower(a.attname)) as name,
                            trim(d.description) as comment
                        from pg_attribute a
                        join pg_description d on
                            d.objoid = a.attrelid and
                            d.objsubid = a.attnum
                        where
                            a.attrelid = c.oid and
                            a.attnum > 0 and
                            not a.attisdropped
                        order by a.attnum asc
                    ) q
                ) as column_comments
            FROM pg_class c
            JOIN pg_namespace n ON
                n.oid = c.relnamespace AND
                n.nspname !~* '^pg_temp' AND
                n.nspname !~* '^pg_toast' AND
                n.nspname != ALL(%s)
            JOIN pg_roles r ON
                r.oid = c.relowner
            WHERE c.relkind in ('v','m')
            ORDER BY 2,3
        """

        query = query.replace("&indexes&", """
                (
                    select jsonb_agg(q order by q.name)
                    from (
                        select
                            trim(lower(ic.relname)) as name,
                            pg_get_indexdef(i.indexrelid, 0, true) as definition
                        from pg_index i
                        join pg_class ic on
                            ic.oid = i.indexrelid
                        where
                            i.indrelid = c.oid and
                            not i.indisprimary and
                            not exists(
                                select 1
                                from pg_constraint co
                                where co.conindid = ic.oid
                            )
                    ) q
                ) as indexes
        """)

        cursor.execute(query, [self.ExcludeSchemas])
        for row in cursor.fetchall():
            self.Objects.update(View(None, row).Export())

        logging.info("GetView loaded = {0}".format(cursor.rowcount))

    @CalcDuration
    def GetSequence(self, cursor):
        cursor.execute("""
            SELECT
                c.oid,
                trim(lower(n.nspname)) AS schema,
                trim(lower(c.relname)) AS name,
                trim(lower(r.rolname)) AS owner,
                trim(coalesce(obj_description(c.oid, 'pg_class'), '')) AS comment,
                c.relacl::varchar[] AS acl,
                s.increment,
                s.minimum_value,
                s.maximum_value,
                s.cycle_option = 'YES' AS is_cycle,
                1 AS start,
                1 AS cache
            FROM pg_class c
            JOIN pg_namespace n ON
                n.oid = c.relnamespace AND
                n.nspname !~* '^pg_temp' AND
                n.nspname !~* '^pg_toast' AND
                n.nspname != ALL(%s)
            JOIN pg_roles r ON
                r.oid = c.relowner
            JOIN information_schema.sequences s ON
                s.sequence_schema = n.nspname and
                s.sequence_name = c.relname
            WHERE c.relkind = 'S'
            ORDER BY 2,3
        """, [self.ExcludeSchemas])

        for row in cursor.fetchall():
            self.Objects.update(Sequence(None, row).Export())

        logging.info("GetSequence loaded = {0}".format(cursor.rowcount))

    @CalcDuration
    def GetForeignServer(self, cursor):
        query = """
            select
                s.oid,
                s.srvname as server_name,
                w.fdwname as fdw_name,
                o.rolname as owner_name,
                s.srvoptions as options,
                s.srvacl::varchar[] AS acl,
                obj_description(s.oid, 'pg_foreign_server') AS comment,
                &user_mappings& as user_mappings
            from pg_foreign_server s
            join pg_roles o on
                o.oid = s.srvowner
            join pg_foreign_data_wrapper w on
                w.oid = s.srvfdw
        """

        if self.IsSuperUser and not Config.IsHideUserMappings:
            query = query.replace("&user_mappings&", """
                (
                    select jsonb_object_agg(coalesce(umr.rolname, 'public'), um.umoptions)
                    from pg_user_mapping um
                    left join pg_roles umr on
                        umr.oid = um.umuser
                    where um.umserver = s.oid
                )""")
        else:
            query = query.replace("&user_mappings&", """
                null::jsonb""")

        cursor.execute(query, [self.ExcludeSchemas])

        for row in cursor.fetchall():
            self.Objects.update(ForeignServer(None, row).Export())

        logging.info("GetForeignServer loaded = {0}".format(cursor.rowcount))

    @CalcDuration
    def GetForeignTable(self, cursor):
        cursor.execute("""
            select
                c.oid,
                n.nspname as schema_name,
                c.relname as table_name,
                o.rolname as owner_name,
                s.srvname as server_name,
                t.ftoptions as options,
                obj_description(c.oid, 'pg_class') AS comment,
                c.relacl::varchar[] AS acl,
                (
                    select array_agg(concat_ws(' ',
                        a.attname,
                        trim(lower(format_type(a.atttypid, a.atttypmod))),
                        case when a.attnotnull then 'NOT NULL' end,
                        case when a.attfdwoptions is not null then format('OPTIONS(%%s)',(
                            select string_agg(distinct concat_ws(' ',
                                q.opt[1], quote_literal(q.opt[2])), ',')
                            from (
                                select regexp_split_to_array(ao, '\=', 'im') opt
                                from unnest(a.attfdwoptions) ao
                            ) q
                        )) end
                    ))
                    from pg_attribute a
                    where
                        a.attrelid = c.oid and
                        a.attnum > 0
                ) as columns_list
            from pg_foreign_table t
            join pg_foreign_server s on
                s.oid = t.ftserver
            join pg_class c on
                c.oid = t.ftrelid
            join pg_roles o on
                o.oid = c.relowner
            join pg_namespace n on
                n.oid = c.relnamespace AND
                n.nspname !~* '^pg_temp' AND
                n.nspname !~* '^pg_toast' AND
                n.nspname != ALL(%s)
        """, [self.ExcludeSchemas])

        for row in cursor.fetchall():
            self.Objects.update(ForeignTable(None, row).Export())

        logging.info("GetForeignTable loaded = {0}".format(cursor.rowcount))

    @CalcDuration
    def GetExtension(self, cursor):
        cursor.execute("""
            SELECT
                e.oid,
                n.nspname AS schema,
                o.rolname AS owner,
                e.extname AS name,
                e.extversion AS version
            FROM pg_extension e
            JOIN pg_namespace n ON
                n.oid = e.extnamespace
            JOIN pg_roles o ON
                o.oid = e.extowner
        """)

        for row in cursor.fetchall():
            self.Objects.update(Extension(None, row).Export())

        logging.info("GetExtension loaded = {0}".format(cursor.rowcount))

    @CalcDuration
    def GetPublication(self, cursor):
        query = """
            select
                p.oid,
                p.pubname as name,
                r.rolname as owner,
                p.pubviaroot as is_via_root,
                p.puballtables as is_all_tables,
                concat_ws(', ',
                    case when p.pubinsert   then 'insert'   end,
                    case when p.pubupdate   then 'update'   end,
                    case when p.pubdelete   then 'delete'   end,
                    case when p.pubtruncate then 'truncate' end
                ) as actions,
                (
                    select array_agg(distinct concat_ws('.', n.nspname, c.relname)
                        order by concat_ws('.', n.nspname, c.relname))
                    from pg_publication_rel pt
                    join pg_class c on
                        c.oid = pt.prrelid
                    join pg_namespace n on
                        n.oid = c.relnamespace
                    where pt.prpubid = p.oid
                ) as tables
            from pg_publication p
            join pg_roles r on
                r.oid = p.pubowner
        """

        if self.Version <= 12:
            query = query.replace("p.pubviaroot", "FALSE")

        cursor.execute(query)
        for row in cursor.fetchall():
            self.Objects.update(Publication(None, row).Export())

        logging.info("GetPublication loaded = {0}".format(cursor.rowcount))

    @CalcDuration
    def GetSubscription(self, cursor):
        if not self.IsSuperUser:
            return

        cursor.execute("""
            select
                s.oid,
                s.subname as name,
                r.rolname as owner,
                s.subenabled as is_enabled,
                (
                    select string_agg(r, ' ')
                    from regexp_split_to_table(s.subconninfo, ' ', 'im') r
                    where r !~* '^password'
                ) as connect,
                s.subslotname as slot,
                s.subsynccommit as sync_commit,
                s.subpublications as publications,
                (
                    select array_agg(distinct concat_ws('.', n.nspname, c.relname)
                        order by concat_ws('.', n.nspname, c.relname))
                    from pg_subscription_rel st
                    join pg_class c on
                        c.oid = st.srrelid
                    join pg_namespace n on
                        n.oid = c.relnamespace
                    where st.srsubid = s.oid
                ) as tables
            from pg_subscription s
            join pg_roles r on
                r.oid = s.subowner
        """)

        for row in cursor.fetchall():
            self.Objects.update(Subscription(None, row).Export())

        logging.info("GetSubscription loaded = {0}".format(cursor.rowcount))

    @CalcDuration
    def GetEventTrigger(self, cursor):
        cursor.execute("""
            select
                t.oid,
                t.evtname as name,
                r.rolname as owner,
                upper(t.evtevent) as event,
                t.evttags as tags,
                format('%s.%s(%s)', n.nspname, p.proname, oidvectortypes(p.proargtypes)) as fnc,
                case t.evtenabled
                    when 'D' then 'DISABLE'
                    when 'R' THEN 'ENABLE REPLICA'
                    when 'A' THEN 'ENABLE ALWAYS'
                    else 'ENABLE'
                end as status
            from pg_event_trigger t
            join pg_roles r on
                r.oid = t.evtowner
            join pg_proc p on
                p.oid = t.evtfoid
            join pg_namespace n on
                n.oid = p.pronamespace
        """)

        for row in cursor.fetchall():
            self.Objects.update(EventTrigger(None, row).Export())

        logging.info("GetEventTrigger loaded = {0}".format(cursor.rowcount))

    @CalcDuration
    def GetCast(self, cursor):
        cursor.execute("""
            select
                c.oid,
                pg_catalog.format_type(ts.oid,ts.typtypmod) as type_from,
                pg_catalog.format_type(tt.oid,tt.typtypmod) as type_to,
                case c.castcontext
                    when 'i' then 'IMPLICIT'
                    when 'a' then 'ASSIGNMENT'
                    when 'e' then 'EXPLICIT'
                end as context,
                case
                    when p.oid is null then null
                    else format('%s.%s(%s)', pn.nspname, p.proname, oidvectortypes(p.proargtypes))
                end as func
            from pg_cast c
            join pg_type ts on
                ts.oid = c.castsource
            join pg_type tt on
                tt.oid = c.casttarget
            left join pg_proc p on
                p.oid = c.castfunc
            left join pg_namespace pn on
                pn.oid = p.pronamespace
            where c.oid > 16383
        """)

        for row in cursor.fetchall():
            self.Objects.update(Cast(None, row).Export())

        logging.info("GetCast loaded = {0}".format(cursor.rowcount))



#!/usr/bin/python
# -*- coding: utf-8 -*-

from postgres_ddl.Comment import Comment
from postgres_ddl.Config import Config
from postgres_ddl.DDL import DDL
from postgres_ddl.Grant import Grant
from postgres_ddl.Owner import Owner
from postgres_ddl.System import ParseACL

class Sequence(DDL):
    def __init__(self, parent, data):
        super().__init__(parent, data)

        self.Oid = data.get("oid")
        assert (self.Oid or 0) > 0, \
            "Sequence oid is null"

        self.Schema = (data.get("schema") or "").strip()
        assert len(self.Schema) > 0, \
            "Sequence schema is null"

        self.Name = (data.get("name") or "").strip()
        assert len(self.Name) > 0, \
            "Sequence name is null"

        self.Increment = int(data.get("increment") or 0)
        assert self.Increment != 0, \
            "Sequence increment is null"

        self.MinValue = int(data.get("minimum_value") or 0)
        self.MaxValue = int(data.get("maximum_value") or 0)

        self.IsCycle = "CYCLE" if data.get("is_cycle") is True else "NO CYCLE"

        self.Cache = (data.get("cache") or 0)
        assert self.Cache > 0, \
            "Sequence cache is null"

        self.Owner = Owner(
            self.GetObjectName(),
            {
                "instance_type" : "TABLE",
                "instance_name" : self.GetFullName(),
                "owner_name"    : data.get("owner")
            }
        )

        self.Comment = Comment(
            self.GetObjectName(),
            {
                "instance_type" : self.GetTag(),
                "instance_name" : self.GetFullName(),
                "comment"       : data.get("comment")
            }
        )

        self.Grants = []
        for grant in ParseACL(data.get("acl"), self.Owner.Owner):
            grant["instance_type"] = self.GetTag()
            grant["instance_name"] = self.GetFullName()
            self.Grants.append(Grant(self.GetObjectName(), grant))

    def __str__(self):
        return self.GetFullName()

    def GetObjectType(self):
        return "sequence"

    def GetObjectName(self):
        return f"{self.GetObjectType()}_{self.GetFullName()}"

    def GetFullName(self):
        return f"{self.Schema}.{self.Name}"

    def GetTag(self):
        return "SEQUENCE"

    def DDL_Drop(self):
        return f"DROP {self.GetTag()} IF EXISTS {self.GetFullName()};"

    def DDL_Create(self):
        r = f"-- Sequence: {self.GetFullName()}"
        r += Config.NL + Config.NL
        r += f"-- {self.DDL_Drop()}"
        r += Config.NL + Config.NL
        r += f"CREATE {self.GetTag()} {self.GetFullName()}"
        r += Config.NL
        r += f"{Config.Indent}INCREMENT {self.Increment}"
        r += Config.NL
        r += f"{Config.Indent}MINVALUE {self.MinValue}"
        r += Config.NL
        r += f"{Config.Indent}MAXVALUE {self.MaxValue}"
        r += Config.NL
        r += f"{Config.Indent}START 1"
        r += Config.NL
        r += f"{Config.Indent}CACHE {self.Cache}"
        r += Config.NL
        r += f"{Config.Indent}{self.IsCycle}"
        r = r.strip() + ";"
        r += Config.NL + Config.NL
        r += self.Owner.DDL_Create()
        r += Config.NL + Config.NL

        for grant in self.Grants:
            r += grant.DDL_Create()
            r += Config.NL
        r += Config.NL

        if self.Comment.IsExists:
            r += self.Comment.DDL_Create()
            r += Config.NL

        return r.strip() + Config.NL

    def GetPath(self):
        return [self.Schema, "sequence"]

    def GetFileName(self):
        return f"{self.Name}.sql"

    def Export(self):
        result = {}
        result[self.GetObjectName()] = self
        result[self.Comment.GetObjectName()] = self.Comment
        result[self.Owner.GetObjectName()] = self.Owner
        for v in self.Grants:
            result[v.GetObjectName()] = v
        return result

    def Diff(self, another):
        result = []

        if self.Cache != another.Cache:
            result.append(f"CACHE {self.Cache}")

        if self.Increment != another.Increment:
            result.append(f"INCREMENT BY {self.Increment}")

        if self.IsCycle != another.IsCycle:
            result.append(self.IsCycle)

        if self.MinValue != another.MinValue:
            result.append(f"MINVALUE {self.MinValue}")

        if self.MaxValue != another.MaxValue:
            result.append(f"MAXVALUE {self.MaxValue}")

        if len(result) > 0:
            result = " ".join(result)
            return [f"ALTER {self.GetTag()} {self.GetFullName()} {result}"]
        else:
            return []



#!/usr/bin/python
# -*- coding: utf-8 -*-

from postgres_ddl.Comment import Comment
from postgres_ddl.Config import Config
from postgres_ddl.DDL import DDL
from postgres_ddl.Grant import Grant
from postgres_ddl.Owner import Owner
from postgres_ddl.System import ParseACL

class TableSequence(DDL):
    def __init__(self, parent, data):
        super().__init__(parent, data)

        self.Schema = (data.get("schema") or "").strip()
        assert len(self.Schema) > 0, \
            "Sequence schema is null"

        self.Name = (data.get("name") or "").strip()
        assert len(self.Name) > 0, \
            "Sequence name is null"

        self.Owner = Owner(
            self.GetObjectName(),
            {
                "instance_type" : "TABLE",
                "instance_name" : self.GetFullName(),
                "owner_name"    : data.get("owner")
            }
        )

        self.Comment = Comment(
            self.GetObjectName(),
            {
                "instance_type" : self.GetTag(),
                "instance_name" : self.GetFullName(),
                "comment"       : data.get("comment")
            }
        )

        self.Grants = []
        for grant in ParseACL(data.get("acl"), self.Owner.Owner):
            grant["instance_type"] = self.GetTag()
            grant["instance_name"] = self.GetFullName()
            self.Grants.append(Grant(self.GetObjectName(), grant))

    def __str__(self):
        return self.GetFullName()

    def GetObjectType(self):
        return "sequence"

    def GetObjectName(self):
        return f"{self.GetObjectType()}_{self.GetFullName()}"

    def GetFullName(self):
        return f"{self.Schema}.{self.Name}"

    def GetTag(self):
        return "SEQUENCE"

    def DDL_Create(self):
        r = f"-- Sequence: {self.GetFullName()}"
        r += Config.NL + Config.NL

        r += self.Owner.DDL_Create()
        r += Config.NL + Config.NL

        for grant in self.Grants:
            r += grant.DDL_Create()
            r += Config.NL
        r += Config.NL

        if self.Comment.IsExists:
            r += self.Comment.DDL_Create()
            r += Config.NL

        return r.strip() + Config.NL



#!/usr/bin/python
# -*- coding: utf-8 -*-

from postgres_ddl.Config import Config
from postgres_ddl.DDL import DDL

class TableConstraint(DDL):
    def __init__(self, parent, data):
        super().__init__(parent, data)

        self.Schema = (data.get("schema") or "").strip().lower()
        assert len(self.Schema) > 0, \
            "Constraint schema is null"

        self.Table = (data.get("table") or "").strip().lower()
        assert len(self.Table) > 0, \
            "Constraint table is null"

        self.Name = (data.get("name") or "").strip().lower()
        assert len(self.Name) > 0, \
            "Constraint name is null"

        self.Type = (data.get("type") or "").strip().lower()
        assert len(self.Type) > 0, \
            "Constraint type is null"

        self.IsForeignKey = (self.Type == "f")

        self.OrderNum = data.get("order_num") or 6

        self.SortKey = f"{self.OrderNum}_{self.GetFullName()}"

        self.Definition = (data.get("definition") or "")
        assert len(self.Definition) > 0, \
            "Constraint definition is null"

        self.UpdateAction = (data.get("update_action") or "").strip().upper()

        self.DeleteAction = (data.get("delete_action") or "").strip().upper()

        self.MatchAction = (data.get("match_action") or "").strip().upper()

        self.DeferrableType = (data.get("deferrable_type") or "").strip().upper()

    def __str__(self):
        return self.GetFullName()

    def GetObjectType(self):
        return "table_constraint"

    def GetObjectName(self):
        return f"{self.GetObjectType()}_{self.GetFullName()}"

    def GetFullName(self):
        return f"{self.Schema}.{self.Table}.{self.Name}"

    def GetTag(self):
        return "CONSTRAINT"

    def GetDefinition(self, separator=" "):
        if not self.IsForeignKey:
            return f"{self.GetTag()} {self.Name} {self.Definition}"

        definition = self.Definition
        if definition.find(" ON") > -1:
            definition = definition[0:definition.find(") ON") + 1]

        #definition = definition.replace("REFERENCES", separator+"REFERENCES")

        return separator.join([
            f"{self.GetTag()} {self.Name}",
            definition,
            self.MatchAction,
            self.UpdateAction,
            self.DeleteAction,
            self.DeferrableType
        ])

    def DDL_Inner(self):
        separator = f"{Config.NL}{Config.Indent * 2}"
        return Config.Indent + self.GetDefinition(separator)

    def DDL_Create(self):
        definition = self.GetDefinition(" ")
        return f"ALTER TABLE {self.Schema}.{self.Table} ADD {definition};"

    def DDL_Drop(self):
        return f"ALTER TABLE {self.Schema}.{self.Table} DROP {self.GetTag()} IF EXISTS {self.Name};"

    def Diff(self, another):
        if (
            self.DeferrableType != another.DeferrableType   or
            self.Definition     != another.Definition       or
            self.UpdateAction   != another.UpdateAction     or
            self.DeleteAction   != another.DeleteAction     or
            self.MatchAction    != another.MatchAction
        ):
            return [
                another.DDL_Drop(),
                self.DDL_Create()
            ]
        else:
            return []



#!/usr/bin/python
# -*- coding: utf-8 -*-

from postgres_ddl.Config import Config
from postgres_ddl.DDL import DDL
from postgres_ddl.Owner import Owner

class Publication(DDL):
    def __init__(self, parent, data):
        super().__init__(parent, data)

        self.Oid = data.get("oid")
        assert (self.Oid or 0) > 0, \
            "Publication oid is null"

        self.Name = (data.get("name") or "").strip()
        assert len(self.Name) > 0, \
            "Publication name is null"

        self.IsViaRoot = bool(data.get("is_via_root") or False)
        self.IsAllTables = bool(data.get("is_all_tables") or False)

        self.Actions = (data.get("actions") or "").strip()
        assert len(self.Actions) > 0, \
            "Publication actions is null"

        self.Tables = (data.get("tables") or [])
        assert self.IsAllTables or len(self.Tables) > 0, \
            "Publication tables is null"

        self.Owner = Owner(
            self.GetObjectName(),
            {
                "instance_type" : self.GetTag(),
                "instance_name" : self.GetFullName(),
                "owner_name"    : data.get("owner")
            }
        )

    def __str__(self):
        return self.GetFullName()

    def GetObjectType(self):
        return "publication"

    def GetObjectName(self):
        return f"{self.GetObjectType()}_{self.GetFullName()}"

    def GetFullName(self):
        return self.Name

    def GetTag(self):
        return "PUBLICATION"

    def DDL_Drop(self):
        return f"DROP {self.GetTag()} IF EXISTS {self.GetFullName()};"

    def DDL_Create(self):
        r = f"-- Publication: {self.GetFullName()}"
        r += Config.NL + Config.NL
        r += f"-- {self.DDL_Drop()}"
        r += Config.NL + Config.NL
        r += f"CREATE {self.GetTag()} {self.GetFullName()}"

        if self.IsAllTables:
            r += f" FOR ALL TABLES"
        else:
            r += f" FOR TABLE"
            r += Config.NL
            r += f",{Config.NL}".join([f"{Config.Indent}{t}" for t in self.Tables])
        r += Config.NL

        r += "WITH ("
        r += Config.NL
        r += f"{Config.Indent}publish = '{self.Actions}',"
        r += Config.NL
        r += f"{Config.Indent}publish_via_partition_root = {str(self.IsViaRoot).upper()}"
        r += Config.NL
        r += ");"

        r += Config.NL + Config.NL
        r += self.Owner.DDL_Create() + Config.NL

        return r.strip() + Config.NL

    def GetPath(self):
        return ["_logical"]

    def GetFileName(self):
        return f"{self.Name}.sql"

    def Export(self):
        result = {}
        result[self.GetObjectName()] = self
        result[self.Owner.GetObjectName()] = self.Owner
        return result

    def Diff(self, another):
        result = []

        if self.IsViaRoot != another.IsViaRoot:
            result.append(f"ALTER {self.GetTag()} {self.GetFullName()} SET(publish_via_partition_root = {str(self.IsViaRoot).upper()});")

        if self.Actions != another.Actions:
            result.append(f"ALTER {self.GetTag()} {self.GetFullName()} SET(publish = '{self.Actions}');")

        if self.Tables != another.Tables:
            tables = ", ".join(self.Tables)
            result.append(f"ALTER {self.GetTag()} {self.GetFullName()} SET TABLE {tables};")

        return result



#!/usr/bin/python
# -*- coding: utf-8 -*-

from postgres_ddl.Config import Config
from postgres_ddl.DDL import DDL

class TableTrigger(DDL):
    def __init__(self, parent, data):
        super().__init__(parent, data)

        self.Schema = (data.get("schema") or "").strip().lower()
        assert len(self.Schema) > 0, \
            "Trigger schema is null"

        self.Table = (data.get("table") or "").strip().lower()
        assert len(self.Table) > 0, \
            "Trigger table is null"

        self.Name = (data.get("name") or "").strip().lower()
        assert len(self.Name) > 0, \
            "Trigger name is null"

        self.IsDisabled = data.get("is_disabled") or False

        self.Definition = data.get("definition") or ""
        assert len(self.Definition) > 0, \
            "Trigger definition is null"

    def __str__(self):
        return self.GetFullName()

    def GetObjectType(self):
        return "table_trigger"

    def GetObjectName(self):
        return f"{self.GetObjectType()}_{self.GetFullName()}"

    def GetFullName(self):
        return f"{self.Schema}.{self.Name}"

    def GetTag(self):
        return "TRIGGER"

    def DDL_Drop(self):
        return f"DROP {self.GetTag()} IF EXISTS {self.Name} ON {self.Schema}.{self.Table};"

    def DDL_Create(self):
        dfn = self.Definition
        dfn = dfn.replace(" BEFORE",  f"{Config.NL}{Config.Indent}BEFORE")
        dfn = dfn.replace(" AFTER",   f"{Config.NL}{Config.Indent}AFTER")
        dfn = dfn.replace(" ON",      f"{Config.NL}{Config.Indent}ON")
        dfn = dfn.replace(" FOR",     f"{Config.NL}{Config.Indent}FOR")
        dfn = dfn.replace(" EXECUTE", f"{Config.NL}{Config.Indent}EXECUTE")

        r = f"-- Trigger: {self.Name} ON {self.Schema}.{self.Table}"
        r += Config.NL + Config.NL
        r += f"-- {self.DDL_Drop()}"
        r += Config.NL + Config.NL
        r += dfn + ";"

        if self.IsDisabled:
            r += Config.NL + Config.NL
            r += self.DDL_Enabled()

        return r.strip()

    def DDL_Enabled(self):
        if self.IsDisabled:
            return f"ALTER TABLE {self.Schema}.{self.Table} DISABLE {self.GetTag()} {self.Name};"
        else:
            return f"ALTER TABLE {self.Schema}.{self.Table} ENABLE {self.GetTag()} {self.Name};"

    def Diff(self, another):
        result = []

        if self.Definition != another.Definition:
            result.append(another.DDL_Drop())
            result.append(self.DDL_Create())

        if self.IsDisabled != another.IsDisabled:
            result.append(self.DDL_Enabled())

        return result



#!/usr/bin/python
# -*- coding: utf-8 -*-

from postgres_ddl.Comment import Comment
from postgres_ddl.Config import Config
from postgres_ddl.DDL import DDL
from postgres_ddl.Grant import Grant
from postgres_ddl.Owner import Owner
from postgres_ddl.System import ParseACL

class ForeignTable(DDL):
    def __init__(self, parent, data):
        super().__init__(parent, data)

        self.Schema = (data.get("schema_name") or "").strip().lower()
        assert len(self.Schema) > 0, \
            "Foreign table schema is null"

        self.Name = (data.get("table_name") or "").strip().lower()
        assert len(self.Name) > 0, \
            "Foreign table name is null"

        self.Server = (data.get("server_name") or "").strip()
        assert len(self.Server) > 0, \
            "Foreign table server is null"

        self.Options = data.get("options")

        self.Columns = data.get("columns_list") or []
        assert len(self.Columns) > 0, \
            "Foreign table columns is null"

        self.Owner = Owner(
            self.GetObjectName(),
            {
                "instance_type" : self.GetTag(),
                "instance_name" : self.GetFullName(),
                "owner_name"    : data.get("owner_name")
            }
        )

        self.Comment = Comment(
            self.GetObjectName(),
            {
                "instance_type" : self.GetTag(),
                "instance_name" : self.GetFullName(),
                "comment"       : data.get("comment")
            }
        )

        self.Grants = []
        for grant in ParseACL(data.get("acl"), self.Owner.Owner):
            grant["instance_type"] = "TABLE"
            grant["instance_name"] = self.GetFullName()
            self.Grants.append(Grant(self.GetObjectName(), grant))

    def __str__(self):
        return self.Name

    def GetObjectType(self):
        return "foreign_server"

    def GetObjectName(self):
        return f"{self.GetObjectType()}_{self.GetFullName()}"

    def GetFullName(self):
        return f"{self.Schema}.{self.Name}"

    def GetTag(self):
        return "FOREIGN TABLE"

    def DDL_Drop(self):
        return f"DROP {self.GetTag()} IF EXISTS {self.GetFullName()};"

    def DDL_Create(self):
        r = f"-- Foreign Table: {self.GetFullName()}"
        r += Config.NL + Config.NL
        r += f"-- {self.DDL_Drop()}"
        r += Config.NL + Config.NL
        r += f"CREATE {self.GetTag()} {self.GetFullName()}("
        r += Config.NL
        r += self.DDL_Columns()
        r += Config.NL
        r += ")"
        r += Config.NL
        r += f"SERVER {self.Server}"
        r += Config.NL
        r += "OPTIONS("
        r += Config.NL
        r += self.DDL_Options()
        r += Config.NL
        r += ");"
        r += Config.NL + Config.NL

        if self.Owner.Owner is not None:
            r += self.Owner.DDL_Create()
            r += Config.NL + Config.NL

        for grant in self.Grants:
            r += grant.DDL_Create()
            r += Config.NL

        if self.Comment.IsExists:
            r += Config.NL
            r += self.Comment.DDL_Create()

        return r.strip() + Config.NL

    def DDL_Options(self):
        result = []

        for o in sorted(self.Options):
            o = o.split("=")
            if len(o) != 2:
                continue
            result.append(f"{Config.Indent}{o[0]} '{o[1]}'")

        return f",{Config.NL}".join(result)

    def DDL_Columns(self):
        result = []
        for col in self.Columns:
            result.append(f"{Config.Indent}{col}")
        return f",{Config.NL}".join(result)

    def GetPath(self):
        return [self.Schema, "foreign_table"]

    def GetFileName(self):
        return f"{self.Name}.sql"

    def Export(self):
        result = {}
        result[self.GetObjectName()] = self
        result[self.Comment.GetObjectName()] = self.Comment
        result[self.Owner.GetObjectName()] = self.Owner
        for v in self.Grants:
            result[v.GetObjectName()] = v
        return result




#!/usr/bin/python
# -*- coding: utf-8 -*-

from postgres_ddl.Config import Config
from postgres_ddl.DDL import DDL

class Owner(DDL):
    def __init__(self, parent, data):
        super().__init__(parent, data)

        self.Type = (data.get("instance_type") or "").strip().upper()
        assert len(self.Type) > 0, \
            f"Owner instance type is null - {parent}"

        self.Instance = (data.get("instance_name") or "").strip()
        assert len(self.Instance) > 0, \
            f"Owner instance name is null - {parent}"

        self.Owner = (data.get("owner_name") or "").strip()
        assert len(self.Owner) > 0, \
            f"Owner name is null - {parent}"

    def __str__(self):
        return self.GetObjectName()

    def GetObjectType(self):
        return "owner"

    def GetObjectName(self):
        return f"{self.GetObjectType()}_{self.Type}_{self.Instance}"

    def DDL_Create(self):
        return f"ALTER {self.Type} {self.Instance}{Config.NL}{Config.Indent}OWNER TO {self.Owner};"

    def DDL_Drop(self):
        return ""

    def Diff(self, another):
        if self.Owner != another.Owner:
            return [
                self.DDL_Create()
            ]
        else:
            return []



#!/usr/bin/python
# -*- coding: utf-8 -*-

import os
import logging
from shutil import rmtree

from postgres_ddl.Database import Database
from postgres_ddl.System import SEP, SetupLogging

class Diff():
    def __init__(self, config):
        self.ConnectSource = config.get("source") or {}
        self.ConnectTarget = config.get("target") or {}

        self.Path = (config.get("path") or "").strip()
        assert len(self.Path) > 0, \
            "Comparer path is null"

        self.PathSource = "/".join([self.Path, "_source"])
        self.PathTarget = "/".join([self.Path, "_target"])
        self.PathDiff   = "/".join([self.Path, "_diff"])

        self.ExcludeSchemas = config.get("exclude_schemas") or []

        SetupLogging(config)

    def GetPath(self, db, root, object):
        # Get file name and path
        file_name = (object.GetFileName() or "").strip()
        path = object.GetPath() or []

        # File name not found - get from parent object
        if len(file_name) <= 0:
            parent = db.Objects.get(object.Parent)
            file_name = (parent.GetFileName() or "").strip()
            path = parent.GetPath() or []

        assert len(file_name) > 0, \
            "Object file name is null"

        assert len(path) > 0, \
            "Object path is null"

        # Create subfolders
        path = "/".join([root] + path)
        if not os.path.exists(path):
            os.makedirs(path)

        return "/".join([path, file_name])

    def Run(self):
        logging.info("Diff started")

        logging.info("Path = {0}".format(self.Path))
        logging.info("PathSource = {0}".format(self.PathSource))
        logging.info("PathTarget = {0}".format(self.PathTarget))
        logging.info("PathDiff = {0}".format(self.PathDiff))

        # Recreate root folder
        if os.path.exists(self.Path):
            rmtree(self.Path)
        os.mkdir(self.Path)

        # Subfolders
        os.mkdir(self.PathSource)
        os.mkdir(self.PathTarget)
        os.mkdir(self.PathDiff)

        logging.info("Folders created")

        # Parse and export source database
        source_db = Database(self.ConnectSource, self.ExcludeSchemas)
        source_db.Parse()

        logging.info("Source db parsed, {0} objects".format(len(source_db.Objects)))

        # Parse and export target database
        target_db = Database(self.ConnectTarget, self.ExcludeSchemas)
        target_db.Parse()

        logging.info("Target db parsed, {0} objects".format(len(source_db.Objects)))

        # Iterate objects of source database
        for sk, sv in source_db.Objects.items():

            # IN source, NOT IN target
            if sk not in target_db.Objects.keys():

                # Parent isn't in target too
                if len(sv.Parent or "") > 0 and sv.Parent not in target_db.Objects.keys():
                    continue

                logging.debug("CREATE {0} not in target".format(sk))

                # Write CREATE to source
                with open(self.GetPath(source_db, self.PathSource, sv), "a", encoding="utf-8") as wf:
                    wf.write(sv.DDL_Create())
                    wf.write(SEP + SEP)

                # Write CREATE to diff
                with open(self.GetPath(source_db, self.PathDiff, sv), "a", encoding="utf-8") as wf:
                    wf.write(sv.DDL_Create())
                    wf.write(SEP + SEP)

                # Write CREATE to object type file
                with open(sv.GetObjectTypeFile(self.Path), "a", encoding="utf-8") as wf:
                    wf.write(tv.DDL_Create())
                    wf.write(SEP + SEP)

            # IN source, IN target
            else:
                tv = target_db.Objects.get(sk)

                # Calculate diff
                diff = sv.Diff(tv)
                if len(diff or []) <= 0:
                    continue

                logging.debug("DIFF {0} both source and target".format(sk))

                # Write CREATE to source
                with open(self.GetPath(source_db, self.PathSource, sv), "a", encoding="utf-8") as wf:
                    wf.write(sv.DDL_Create())
                    wf.write(SEP + SEP)

                # Write CREATE to target
                with open(self.GetPath(target_db, self.PathTarget, tv), "a", encoding="utf-8") as wf:
                    wf.write(tv.DDL_Create())
                    wf.write(SEP + SEP)

                # Write DROP and CREATE to diff
                with open(self.GetPath(source_db, self.PathDiff, sv), "a", encoding="utf-8") as wf:
                    wf.write((SEP+SEP).join(diff))
                    wf.write(SEP + SEP)

                # Write DROP and CREATE to object type file
                with open(sv.GetObjectTypeFile(self.Path), "a", encoding="utf-8") as wf:
                    wf.write((SEP).join(diff))
                    wf.write(SEP + SEP)

        # Iterate objects of target database
        for tk, tv in target_db.Objects.items():

            # NOT IN source, IN target
            if tk in source_db.Objects.keys():
                continue

            # Parent object deleted in source DB
            if len(tv.Parent or "") > 0 and tv.Parent not in source_db.Objects.keys():
                continue

            logging.debug("DROP {0} not in source".format(sk))

            # Write CREATE to target
            with open(self.GetPath(target_db, self.PathTarget, tv), "a", encoding="utf-8") as wf:
                wf.write(tv.DDL_Create())
                wf.write(SEP + SEP)

            # Write DROP to diff
            with open(self.GetPath(target_db, self.PathDiff, tv), "a", encoding="utf-8") as wf:
                wf.write(tv.DDL_Drop())
                wf.write(SEP + SEP)

            # Write DROP to object type file
            with open(sv.GetObjectTypeFile(self.Path), "a", encoding="utf-8") as wf:
                wf.write(tv.DDL_Drop())
                wf.write(SEP + SEP)



#!/usr/bin/python
# -*- coding: utf-8 -*-

class DDL(object):
    def __init__(self, parent, data):
        self.Parent = (parent or "").strip().lower()

        assert data is not None, \
            "{0} data is null".format(type(self))

        assert isinstance(data, dict), \
            "{0} data type is not dict".format(type(self))

        assert len(data.keys()) > 0, \
            "{0} data is empty".format(type(self))

    def __str__(self):
        raise NotImplementedError("__str__ method is not implemented for class {0}".format(type(self)))

    def DDL_Create(self):
        raise NotImplementedError("DDL_Create method is not implemented for class {0}".format(type(self)))

    def DDL_Drop(self):
        raise NotImplementedError("DDL_Drop method is not implemented for class {0}".format(type(self)))

    def GetObjectType(self):
        raise NotImplementedError("GetObjectType method is not implemented for class {0}".format(type(self)))

    def GetObjectName(self):
        raise NotImplementedError("GetObjectName method is not implemented for class {0}".format(type(self)))

    def GetObjectTypeFile(self, path):
        return "/".join([path, "{0}.sql".format(self.GetObjectType())])

    def GetFullName(self):
        raise NotImplementedError("GetFullName method is not implemented for class {0}".format(type(self)))

    def GetTag(self):
        raise NotImplementedError("GetTag method is not implemented for class {0}".format(type(self)))

    def Export(self):
        raise NotImplementedError("Export method is not implemented for class {0}".format(type(self)))

    def GetPath(self):
        return []

    def GetFileName(self):
        return None

    def GetChildObjects(self):
        return {}

    def Diff(self, another):
        return []



#!/usr/bin/python
# -*- coding: utf-8 -*-

from postgres_ddl.Config import Config
from postgres_ddl.DDL import DDL

class TableSettings(DDL):
    def __init__(self, parent, data):
        super().__init__(parent, data)

        self.Schema = (data.get("schema") or "").strip().lower()
        assert len(self.Schema) > 0, \
            "Table settings schema is null"

        self.Table = (data.get("table") or "").strip().lower()
        assert len(self.Table) > 0, \
            "Table settings table is null"

        setting = data.get("setting" or "").strip().lower()
        assert len(setting) > 0, \
            "Table settings value is null"
        assert setting.find('=') >= 0, \
            "Table settings value isn't contains '=' symbol"

        self.Field = setting.split('=')[0].strip().upper()
        self.Value = setting.split('=')[1].strip().upper()

    def __str__(self):
        return f"{self.Field}={self.Value}"

    def GetObjectType(self):
        return "table_setting"

    def GetObjectName(self):
        return f"{self.GetObjectType()}_{self.GetFullName()}"

    def GetFullName(self):
        return f"{self.Schema}.{self.Table}.{self.Field}"

    def DDL_Inner(self):
        return f"{Config.Indent}{self.Field}={self.Value}"

    def DDL_Drop(self):
        return f"ALTER TABLE {self.Schema}.{self.Table} RESET ({self.Field});"

    def DDL_Create(self):
        return f"ALTER TABLE {self.Schema}.{self.Table} SET ({self.Field} = {self.Value});"

    def Diff(self, another):
        if self.Value != another.Value:
            return [
                self.DDL_Create()
            ]
        else:
            return []



#!/usr/bin/python
# -*- coding: utf-8 -*-

from postgres_ddl.Comment import Comment
from postgres_ddl.Config import Config
from postgres_ddl.DDL import DDL
from postgres_ddl.Grant import Grant
from postgres_ddl.Owner import Owner
from postgres_ddl.System import ParseACL

class Function(DDL):
    def __init__(self, parent, data):
        super().__init__(parent, data)

        self.Oid = (data.get("oid") or 0)
        assert self.Oid > 0, \
            "Function oid is null"

        self.Schema = (data.get("schema") or "").strip()
        assert len(self.Schema) > 0, \
            "Function schema is null"

        self.Name = (data.get("proc") or "").strip()
        assert len(self.Name) > 0, \
            "Function name is null"

        self.ArgsInTypes = (data.get("args_in_types") or "").strip()

        self.NameWithParams = f"{self.Schema}.{self.Name}({self.ArgsInTypes})"

        self.ArgsIn = (data.get("args_in") or "").strip()

        self.ArgsOut = (data.get("args_out") or "").strip()

        self.Cost = data.get("cost") or 0

        self.Rows = data.get("rows") or 0

        self.Language = (data.get("lang") or "").strip()
        assert len(self.Language) > 0, \
            "Function language is null"

        self.Volatility = (data.get("volatility") or "").strip()
        assert len(self.Volatility) > 0, \
            "Function volatility"

        self.HasDuplicate = data.get("has_duplicate") or False
        self.IsTrigger = data.get("is_trigger") or False

        self.Code = (data.get("code") or "").strip()
        assert len(self.Code) > 0, \
            "Function code is null"

        self.Owner = Owner(
            self.GetObjectName(),
            {
                "instance_type" : self.GetTag(),
                "instance_name" : self.NameWithParams,
                "owner_name"    : data.get("owner")
            }
        )

        self.Comment = Comment(
            self.GetObjectName(),
            {
                "instance_type" : self.GetTag(),
                "instance_name" : self.NameWithParams,
                "comment"       : data.get("comment")
            }
        )

        self.Grants = []
        for grant in ParseACL(data.get("acl"), self.Owner.Owner):
            grant["instance_type"] = self.GetTag()
            grant["instance_name"] = self.NameWithParams
            self.Grants.append(Grant(self.GetObjectName(), grant))

    def __str__(self):
        return self.GetFullName()

    def GetObjectType(self):
        return "function"

    def GetObjectName(self):
        return f"{self.GetObjectType()}_{self.GetFullName()}"

    def GetFullName(self):
        if self.HasDuplicate:
            return f"{self.Schema}.{self.Name}({self.ArgsInTypes}).sql"
        else:
            return f"{self.Schema}.{self.Name}"

    def GetTag(self):
        return "FUNCTION"

    def DDL_Drop(self):
        return f"DROP {self.GetTag()} IF EXISTS {self.Schema}.{self.Name}({self.ArgsInTypes});"

    def DDL_ArgsIn(self):
        if len(self.ArgsIn or "") == 0:
            return "()" + Config.NL
        else:
            args = self.ArgsIn.replace(",", f",{Config.NL}   ")
            r = ""
            r += "("
            r += Config.NL
            r += f"    {args}"
            r += Config.NL
            r += ")"
            r += Config.NL
            return r

    def DDL_Create(self):
        r = f"-- Function: {self.Schema}.{self.Name}({self.ArgsInTypes})"
        r += Config.NL + Config.NL
        r += f"-- {self.DDL_Drop()}"
        r += Config.NL + Config.NL
        r += f"CREATE OR REPLACE {self.GetTag()} {self.Schema}.{self.Name}"
        r += self.DDL_ArgsIn()
        r += f"RETURNS {self.ArgsOut} AS"
        r += Config.NL
        r += "$BODY$"
        r += Config.NL
        r += self.Code
        r += Config.NL
        r += "$BODY$"
        r += Config.NL
        r += f"{Config.Indent}LANGUAGE {self.Language} {self.Volatility}"

        if self.Cost > 0:
            r += Config.NL
            r += f"{Config.Indent}COST {int(self.Cost)}"

        if self.Rows > 0:
            r += Config.NL
            r += f"{Config.Indent}ROWS {int(self.Rows)}"

        r += ";"
        r += Config.NL + Config.NL
        r += self.Owner.DDL_Create()
        r += Config.NL + Config.NL

        for grant in self.Grants:
            r += grant.DDL_Create()
            r += Config.NL
        r += Config.NL

        if self.Comment.IsExists:
            r += self.Comment.DDL_Create()
            r += Config.NL

        return r.strip() + Config.NL

    def GetPath(self):
        return [self.Schema, "triggers" if self.IsTrigger else "functions"]

    def GetFileName(self):
        if self.HasDuplicate:
            return f"{self.Name}({self.ArgsInTypes}).sql"
        else:
            return f"{self.Name}.sql"

    def Export(self):
        result = {}
        result[self.GetObjectName()] = self
        result[self.Comment.GetObjectName()] = self.Comment
        result[self.Owner.GetObjectName()] = self.Owner
        for g in self.Grants:
            result[g.GetObjectName()] = g
        return result

    def Diff(self, another):
        if (
            self.ArgsIn     != another.ArgsIn     or
            self.ArgsOut    != another.ArgsOut    or
            self.Code       != another.Code       or
            self.Cost       != another.Cost       or
            self.Language   != another.Language   or
            self.Rows       != another.Rows       or
            self.Volatility != another.Volatility
        ):
            return [
                another.DDL_Drop(),
                self.DDL_Create()
            ]
        else:
            return []



#!/usr/bin/python
# -*- coding: utf-8 -*-

from postgres_ddl.Config import Config
from postgres_ddl.DDL import DDL
from postgres_ddl.Owner import Owner

class EventTrigger(DDL):
    def __init__(self, parent, data):
        super().__init__(parent, data)

        self.Oid = data.get("oid")
        assert (self.Oid or 0) > 0, \
            "Event trigger oid is null"

        self.Name = (data.get("name") or "").strip()
        assert len(self.Name) > 0, \
            "Event trigger name is null"

        self.Status = (data.get("status") or "").strip()
        assert len(self.Status) > 0, \
            "Event trigger status is null"

        self.Event = (data.get("event") or "").strip()
        assert len(self.Event) > 0, \
            "Event trigger event is null"

        self.Function = (data.get("fnc") or "").strip()
        assert len(self.Function) > 0, \
            "Event trigger function is null"

        self.Tags = (data.get("tags") or [])

        self.Owner = Owner(
            self.GetObjectName(),
            {
                "instance_type" : self.GetTag(),
                "instance_name" : self.GetFullName(),
                "owner_name"    : data.get("owner")
            }
        )

    def __str__(self):
        return self.GetFullName()

    def GetObjectType(self):
        return "event_trigger"

    def GetObjectName(self):
        return f"{self.GetObjectType()}_{self.GetFullName()}"

    def GetFullName(self):
        return self.Name

    def GetTag(self):
        return "EVENT TRIGGER"

    def DDL_Drop(self):
        return f"DROP {self.GetTag()} IF EXISTS {self.GetFullName()};"

    def DDL_Create(self):
        r = f"-- Event trigger: {self.GetFullName()}"
        r += Config.NL + Config.NL
        r += f"-- {self.DDL_Drop()}"
        r += Config.NL + Config.NL
        r += f"CREATE {self.GetTag()} {self.GetFullName()}"
        r += Config.NL
        r += f"{Config.Indent}ON {self.Event}"
        r += Config.NL

        if len(self.Tags) > 0:
            tags = ", ".join([f"'{i}'" for i in sorted(self.Tags)])
            r += f"{Config.Indent}WHEN TAG IN ({tags})"
            r += Config.NL

        r += f"{Config.Indent}EXECUTE PROCEDURE {self.Function};"
        r += Config.NL + Config.NL
        r += self.Owner.DDL_Create()
        r += Config.NL + Config.NL
        r += self.DDL_Status()
        return r.strip() + Config.NL

    def DDL_Status(self):
        return f"ALTER {self.GetTag()} {self.GetFullName()} {self.Status};"

    def GetPath(self):
        return ["_event_trigger"]

    def GetFileName(self):
        return f"{self.Name}.sql"

    def Export(self):
        result = {}
        result[self.GetObjectName()] = self
        result[self.Owner.GetObjectName()] = self.Owner
        return result

    def Diff(self, another):
        if (
            self.Event != another.Event or
            self.Function != another.Function or
            self.Tags != another.Tags
        ):
            return [self.DDL_Drop(), self.DDL_Create()]

        if self.Status != another.Status:
            return [self.DDL_Status()]



#!/usr/bin/python
# -*- coding: utf-8 -*-

from postgres_ddl.Comment import Comment
from postgres_ddl.Config import Config
from postgres_ddl.DDL import DDL
from postgres_ddl.Grant import Grant
from postgres_ddl.Owner import Owner
from postgres_ddl.System import ParseACL
from postgres_ddl.TableColumn import TableColumn
from postgres_ddl.TableConstraint import TableConstraint
from postgres_ddl.TableIndex import TableIndex
from postgres_ddl.TablePolicy import TablePolicy
from postgres_ddl.TableSequence import TableSequence
from postgres_ddl.TableSettings import TableSettings
from postgres_ddl.TableTrigger import TableTrigger

class Table(DDL):
    def __init__(self, parent, data):
        super().__init__(parent, data)

        self.Oid = (data.get("oid") or 0)
        assert self.Oid > 0, \
            "Table oid is null"

        self.Schema = (data.get("schema") or "").strip()
        assert len(self.Schema) > 0, \
            "Table schema is null"

        self.Name = (data.get("name") or "").strip()
        assert len(self.Name) > 0, \
            "Table name is null"

        self.Settings = []
        for o in (data.get("reloptions") or []):
            self.Settings.append(TableSettings(self.GetObjectName(), {
                "schema"    : self.Schema,
                "table"     : self.Name,
                "setting"   : o
            }))

        self.HasOids = (data.get("has_oids") or "").strip()
        assert len(self.HasOids) > 0, \
            "Table has oids is null"

        self.HasOids = f"OIDS={str(self.HasOids).upper()}"
        self.Settings.append(TableSettings(self.GetObjectName(), {
            "schema"    : self.Schema,
            "table"     : self.Name,
            "setting"   : self.HasOids
        }))

        self.Inherits = (data.get("parent_table") or "").strip()

        self.Inherits = data.get("parent_table")
        if self.Inherits is not None and self.Inherits != "":
            self.Inherits = self.Inherits.strip()
        else:
            self.Inherits = None

        self.PartKey = data.get("part_key")
        if self.PartKey is not None and self.PartKey != "":
            self.PartKey = self.PartKey.strip()
        else:
            self.PartKey = None

        self.PartBorder = data.get("part_border")

        self.Owner = Owner(
            self.GetObjectName(),
            {
                "instance_type" : self.GetTag(),
                "instance_name" : self.GetFullName(),
                "owner_name"    : data.get("owner")
            }
        )

        self.Comment = Comment(
            self.GetObjectName(),
            {
                "instance_type" : self.GetTag(),
                "instance_name" : self.GetFullName(),
                "comment"       : data.get("comment")
            }
        )

        self.Grants = []
        for grant in ParseACL(data.get("acl"), self.Owner.Owner):
            grant["instance_type"] = self.GetTag()
            grant["instance_name"] = self.GetFullName()
            self.Grants.append(Grant(self.GetObjectName(), grant))

        self.Columns = []
        for column in (data.get("columns") or []):
            column["schema"] = self.Schema
            column["table"] = self.Name
            column = TableColumn(self.GetObjectName(), column)
            self.Columns.append(column)

        self.Constraints = []
        for cons in (data.get("constraints") or []):
            cons["schema"] = self.Schema
            cons["table"] = self.Name
            cons = TableConstraint(self.GetObjectName(), cons)
            self.Constraints.append(cons)

        self.Indexes = []
        for idx in (data.get("indexes") or []):
            idx["schema"] = self.Schema
            idx["table"] = self.Name
            idx = TableIndex(self.GetObjectName(), idx)
            self.Indexes.append(idx)

        self.Triggers = []
        for trg in (data.get("triggers") or []):
            trg["schema"] = self.Schema
            trg["table"] = self.Name
            trg = TableTrigger(self.GetObjectName(), trg)
            self.Triggers.append(trg)

        self.Sequence = None
        if data.get("sequence") is not None:
            self.Sequence = TableSequence(self.GetFullName(), data.get("sequence") or [])

        self.Policies = []
        for plc in (data.get("policies") or []):
            plc["schema"] = self.Schema
            plc["table"] = self.Name
            plc = TablePolicy(self.GetObjectName(), plc)
            self.Policies.append(plc)

    def __str__(self):
        return self.GetFullName()

    def GetObjectType(self):
        return "table"

    def GetObjectName(self):
        return f"{self.GetObjectType()}_{self.GetFullName()}"

    def GetFullName(self):
        return f"{self.Schema}.{self.Name}"

    def GetTag(self):
        return "TABLE"

    def DDL_Drop(self):
        return f"DROP {self.GetTag()} IF EXISTS {self.GetFullName()};"

    def DDL_Settings(self):
        r = []

        for sts in sorted(self.Settings, key=lambda x: x.GetFullName()):
            r.append(sts.DDL_Inner())

        return (f",{Config.NL}").join(r)

    def DDL_Create(self):
        is_last_comma = len(self.Constraints) > 0

        r = f"-- Table: {self.GetFullName()}"
        r += Config.NL + Config.NL
        r += f"-- {self.DDL_Drop()}"
        r += Config.NL + Config.NL
        r += f"CREATE {self.GetTag()} {self.GetFullName()}("
        r += Config.NL

        for col in sorted(self.Columns, key=lambda x: x.OrderNum):
            r += col.DDL_Inner(add_comma=is_last_comma, add_comment=True)
            r += Config.NL

        self.Constraints = sorted(self.Constraints, key=lambda x: x.SortKey)
        if len(self.Constraints) > 0:
            cns = [c.DDL_Inner() for c in self.Constraints]
            cns_NL = f",{Config.NL}"
            r += cns_NL.join(cns)
            r += Config.NL

        r += ")"
        r += Config.NL

        if self.PartKey is not None:
            r += f"PARTITION BY {self.PartKey}"
            r += Config.NL

        if self.Inherits is not None:
            if self.PartBorder is not None:
                r += f"PARTITION OF {self.Inherits}"
                r += Config.NL
            else:
                r += f"INHERITS ( {self.Inherits} )"
                r += Config.NL

        if self.PartBorder is not None:
            r += self.PartBorder
            r += Config.NL

        r += "WITH ("
        r += Config.NL
        r += self.DDL_Settings()
        r += Config.NL
        r += ");"
        r += Config.NL + Config.NL
        r += self.Owner.DDL_Create()
        r += Config.NL

        if len(self.Grants) > 0:
            r += Config.NL
            for grant in self.Grants:
                r += grant.DDL_Create()
                r += Config.NL
        r += Config.NL

        if self.Comment.IsExists:
            r += self.Comment.DDL_Create()
            r += Config.NL + Config.NL

        for col in self.Columns:
            cmt = col.DDL_Comment()
            if cmt is not None:
                r += col.DDL_Comment()
                r += Config.NL
        r += Config.NL

        if Config.IsSeqGrantsInTable and self.Sequence is not None:
            r += self.Sequence.DDL_Create()
            r += Config.NL

        for ind in sorted(self.Indexes, key=lambda x: x.GetFullName()):
            r += ind.DDL_Create(f"{Config.NL}{Config.Indent}")
            r += Config.NL + Config.NL

        for trg in sorted(self.Triggers, key=lambda x: x.GetFullName()):
            r += trg.DDL_Create()
            r += Config.NL + Config.NL

        for plc in sorted(self.Policies, key=lambda x: x.GetFullName()):
            r += plc.DDL_Create()
            r += Config.NL + Config.NL

        return r.strip() + Config.NL

    def GetPath(self):
        return [self.Schema, "table"]

    def GetFileName(self):
        return f"{self.Name}.sql"

    def Export(self):
        result = {}
        result[self.GetObjectName()] = self
        result[self.Comment.GetObjectName()] = self.Comment
        result[self.Owner.GetObjectName()] = self.Owner
        for v in self.Grants:
            result[v.GetObjectName()] = v
        for v in self.Columns:
            result[v.GetObjectName()] = v
        for v in self.Settings:
            result[v.GetObjectName()] = v
        for v in self.Constraints:
            result[v.GetObjectName()] = v
        for v in self.Indexes:
            result[v.GetObjectName()] = v
        for v in self.Triggers:
            result[v.GetObjectName()] = v
        for v in self.Policies:
            result[v.GetObjectName()] = v
        return result



#!/usr/bin/python
# -*- coding: utf-8 -*-

from postgres_ddl.Comment import Comment
from postgres_ddl.Config import Config
from postgres_ddl.DDL import DDL
from postgres_ddl.Grant import Grant
from postgres_ddl.Owner import Owner
from postgres_ddl.System import ParseACL
from postgres_ddl.TableIndex import TableIndex

class View(DDL):
    def __init__(self, parent, data):
        super().__init__(parent, data)

        self.IsMaterialized = data.get("is_materialized")
        self.IsMaterialized = self.IsMaterialized if self.IsMaterialized is not None else False

        self.Oid = data.get("oid")
        assert (self.Oid or 0) > 0, \
            "View oid is null"

        self.Schema = (data.get("schema") or "").strip().lower()
        assert len(self.Schema) > 0, \
            "View schema is null"

        self.Name = (data.get("name") or "").strip().lower()
        assert len(self.Name) > 0, \
            "View name is null"

        self.Definition = (data.get("definition") or "").strip()
        assert len(self.Definition) > 0, \
            "View definition is null"

        self.Owner = Owner(
            self.GetObjectName(),
            {
                "instance_type" : "TABLE",
                "instance_name" : self.GetFullName(),
                "owner_name"    : data.get("owner_name")
            }
        )

        self.Comment = Comment(
            self.GetObjectName(),
            {
                "instance_type" : self.GetTag(),
                "instance_name" : self.GetFullName(),
                "comment"       : data.get("comment")
            }
        )

        self.ColumnComments = []
        for comment in data.get("column_comments") or []:
            comment["instance_type"] = "COLUMN"
            comment["instance_name"] = ".".join([self.GetFullName(), comment.get("name")])
            self.ColumnComments.append(Comment(self.GetObjectName(), comment))

        self.Grants = []
        for grant in ParseACL(data.get("acl"), self.Owner.Owner):
            grant["instance_type"] = "TABLE"
            grant["instance_name"] = self.GetFullName()
            self.Grants.append(Grant(self.GetObjectName(), grant))

        self.Indexes = []
        for idx in (data.get("indexes") or []):
            idx["schema"] = self.Schema
            idx["table"] = self.Name
            idx = TableIndex(self.GetObjectName(), idx)
            self.Indexes.append(idx)

    def __str__(self):
        return self.GetFullName()

    def GetObjectType(self):
        if self.IsMaterialized:
            return "matview"
        else:
            return "view"

    def GetObjectName(self):
        return f"{self.GetObjectType()}_{self.Schema}.{self.Name}"

    def GetFullName(self):
        return f"{self.Schema}.{self.Name}"

    def GetTag(self):
        return "MATERIALIZED VIEW" if self.IsMaterialized else "VIEW"

    def DDL_Drop(self):
        return f"DROP {self.GetTag()} IF EXISTS {self.GetFullName()};"

    def DDL_Create(self):
        r = f"-- View: {self.GetFullName()}"
        r += Config.NL + Config.NL
        r += f"-- {self.DDL_Drop()}"
        r += Config.NL + Config.NL

        if self.IsMaterialized:
            r += f"CREATE {self.GetTag()} {self.GetFullName()} AS"
        else:
            r += f"CREATE OR REPLACE {self.GetTag()} {self.GetFullName()} AS"
        r += Config.NL
        r += self.Definition
        r += Config.NL + Config.NL
        r += self.Owner.DDL_Create()
        r += Config.NL + Config.NL

        for grant in self.Grants:
            r += grant.DDL_Create()
            r += Config.NL
        r += Config.NL

        if self.Comment.IsExists:
            r += self.Comment.DDL_Create()
            r += Config.NL + Config.NL

        for comment in self.ColumnComments:
            r += comment.DDL_Create()
            r += Config.NL + Config.NL

        for ind in sorted(self.Indexes, key=lambda x: x.GetFullName()):
            r += ind.DDL_Create(f"{Config.NL}{Config.Indent}")
            r += Config.NL + Config.NL

        return r.strip() + Config.NL

    def GetPath(self):
        return [self.Schema, "view"]

    def GetFileName(self):
        return f"{self.Name}.sql"

    def Export(self):
        result = {}
        result[self.GetObjectName()] = self
        result[self.Comment.GetObjectName()] = self.Comment
        result[self.Owner.GetObjectName()] = self.Owner
        for v in self.Grants:
            result[v.GetObjectName()] = v
        for v in self.Indexes:
            result[v.GetObjectName()] = v
        return result

    def Diff(self, another):
        if self.Definition != another.Definition:
            return [
                another.DDL_Drop(),
                self.DDL_Create()
            ]




#!/usr/bin/python
# -*- coding: utf-8 -*-

from postgres_ddl.DDL import DDL
from postgres_ddl.Config import Config
from postgres_ddl.System import ParseOptions

class UserMapping(DDL):
    def __init__(self, parent, data):
        super().__init__(parent, data)

        self.Type = (data.get("instance_type") or "").strip().upper()
        assert len(self.Type) > 0, \
            f"UserMapping instance type is null - {parent}"

        self.Instance = (data.get("instance_name") or "").strip()
        assert len(self.Instance) > 0, \
            f"UserMapping instance name is null - {parent}"

        self.Role = (data.get("role") or "").strip().lower()

        self.Options = sorted(data.get("options") or [])

    def __str__(self):
        return self.GetObjectName()

    def GetObjectType(self):
        return "user_mapping"

    def GetObjectName(self):
        return f"{self.GetObjectType()}_{self.Instance}_{self.Role}"

    def GetTag(self):
        return "USER MAPPING"

    def DDL_Create(self):
        return f"CREATE {self.GetTag()} FOR {self.Role} {self.Type} {self.Instance} OPTIONS({self.DDL_Options()});"

    def DDL_Drop(self):
        return f"DROP {self.GetTag()} IF EXISTS FOR {self.Role} {self.Type} {self.Instance};"

    def DDL_Options(self):
        result = []

        for o in self.Options:
            o = ParseOptions(o, ["password"])
            if o is not None:
                result.append(o)

        return f",{Config.NL}".join(sorted(result))

    def Diff(self, another):
        if self.Options != another.Options:
            return [
                another.DDL_Drop(),
                self.DDL_Create()
            ]
        else:
            return []



#!/usr/bin/python
# -*- coding: utf-8 -*-

import re
from postgres_ddl.Config import Config
from postgres_ddl.DDL import DDL

class TableColumn(DDL):
    def __init__(self, parent, data):
        super().__init__(parent, data)

        self.Schema = (data.get("schema") or "").strip().lower()
        assert len(self.Schema) > 0, \
            "Column schema is null"

        self.Table = (data.get("table") or "").strip().lower()
        assert len(self.Table) > 0, \
            "Column table is null"

        self.Name = (data.get("name") or "").strip().lower()
        assert len(self.Name) > 0, \
            "Column name is null"

        self.Type = (data.get("type") or "").strip()
        assert len(self.Type) > 0, \
            "Column type is null"

        self.NotNull = data.get("not_null") or False

        self.DefaultValue = (data.get("default_value") or "").strip()

        self.OrderNum = data.get("order_num") or 0
        assert self.OrderNum > 0, \
            "Column order num is null"

        self.OrderNumLast = data.get("max_order_num") or 0
        assert self.OrderNumLast > 0, \
            "Column max order num is null"

        self.IsLast = (self.OrderNum == self.OrderNumLast)

        self.Comment = data.get("comment")
        if self.Comment is not None:
            self.Comment = re.sub("[\n\r]{1,}", Config.NL, self.Comment, re.MULTILINE | re.IGNORECASE)

    def __str__(self):
        return self.GetFullName()

    def GetObjectType(self):
        return "table_column"

    def GetObjectName(self):
        return f"{self.GetObjectType()}_{self.GetFullName()}"

    def GetFullName(self):
        return f"{self.Schema}.{self.Table}.{self.Name}"

    def GetTag(self):
        return "COLUMN"

    def DDL_Create(self):
        not_null = "" if not self.NotNull else " NOT NULL"
        dflt = "" if len(self.DefaultValue) <= 0 else " DEFAULT " + self.DefaultValue

        r = f"ALTER TABLE {self.Schema}.{self.Table} ADD {self.GetTag()} {self.Name} {self.Type}{not_null}{dflt};"

        if self.Comment is not None and self.Comment.strip() != "":
            r += Config.NL
            r += self.DDL_Comment()

        return r

    def DDL_NotNull(self):
        if self.NotNull:
            return f"ALTER TABLE {self.Schema}.{self.Table} ALTER {self.GetTag()} {self.Name} SET NOT NULL;"
        else:
            return f"ALTER TABLE {self.Schema}.{self.Table} ALTER {self.GetTag()} {self.Name} DROP NOT NULL;"

    def DDL_Default(self):
        if self.DefaultValue is None or self.DefaultValue.strip() == "":
            return f"ALTER TABLE {self.Schema}.{self.Table} ALTER {self.GetTag()} {self.Name} DROP DEFAULT;"
        else:
            return f"ALTER TABLE {self.Schema}.{self.Table} ALTER {self.GetTag()} {self.Name} SET DEFAULT {self.DefaultValue};"

    def DDL_Drop(self):
        return f"ALTER TABLE {self.Schema}.{self.Table} DROP {self.GetTag()} IF EXISTS {self.Name};"

    def DDL_Comment(self):
        if self.Comment is not None and self.Comment.strip() != "":
            return f"COMMENT ON {self.GetTag()} {self.Schema}.{self.Table}.{self.Name}{Config.NL}{Config.Indent}IS '{self.Comment}';"
        return None

    def DDL_Type(self):
        return f"ALTER TABLE {self.Schema}.{self.Table} ALTER {self.GetTag()} {self.Name} TYPE {self.Type};"

    def DDL_Inner(self, add_comma=False, add_comment=False):
        if not add_comma:
            add_comma = not self.IsLast

        return f"{Config.Indent}%s %s%s%s%s%s" % (
            self.Name,
            self.Type,
            "" if not self.NotNull else " NOT NULL",
            "" if len(self.DefaultValue) <= 0 else " DEFAULT " + self.DefaultValue,
            "" if not add_comma else ",",
            "" if not add_comment or self.Comment is None else f" -- %s" % (self.Comment.replace(Config.NL, " "))
        )

    def Diff(self, another):
        result = []

        if self.Comment != another.Comment:
            result.append(self.DDL_Comment())

        if self.DefaultValue != another.DefaultValue:
            result.append(self.DDL_Default())

        if self.NotNull != another.NotNull:
            result.append(self.DDL_NotNull())

        if self.Type != another.Type:
            result.append(self.DDL_Type())

        return result



#!/usr/bin/python
# -*- coding: utf-8 -*-

from postgres_ddl.Config import Config
from postgres_ddl.DDL import DDL

class TablePolicy(DDL):
    def __init__(self, parent, data):
        super().__init__(parent, data)

        self.Schema = (data.get("schema") or "").strip().lower()
        assert len(self.Schema) > 0, \
            "Policy schema is null"

        self.Table = (data.get("table") or "").strip().lower()
        assert len(self.Table) > 0, \
            "Policy table is null"

        self.Name = (data.get("name") or "").strip().lower()
        assert len(self.Name) > 0, \
            "Policy name is null"

        self.Command = (data.get("command") or "").strip().upper()
        assert len(self.Command) > 0, \
            "Policy command is null"

        self.Type = (data.get("type") or "").strip().upper()
        assert len(self.Type) > 0, \
            "Policy type is null"

        self.ExprUsing = data.get("expr_using") or ""
        self.ExporWithCheck = data.get("expr_with_check") or ""
        assert (len(self.ExprUsing) > 0 or len(self.ExporWithCheck) > 0), \
            "Policy expression is null"

        self.Roles = (data.get("roles") or [])
        assert len(self.Roles) > 0, \
            "Policy roles is empty"
        self.Roles = ", ".join(self.Roles)

    def __str__(self):
        return self.GetFullName()

    def GetObjectType(self):
        return "table_policy"

    def GetObjectName(self):
        return f"{self.GetObjectType()}_{self.GetFullName()}"

    def GetFullName(self):
        return f"{self.Schema}.{self.Table}.{self.Name}"

    def GetTag(self):
        return "POLICY"

    def DDL_Drop(self):
        return f"DROP {self.GetTag()} IF EXISTS {self.Name} ON {self.Schema}.{self.Table};"

    def DDL_Create(self):
        r = f"-- Policy: {self.Name} ON {self.Schema}.{self.Table}"
        r += Config.NL + Config.NL
        r += f"-- {self.DDL_Drop()}"
        r += Config.NL + Config.NL
        r += f"CREATE {self.GetTag()} {self.Name}"
        r += Config.NL
        r += f"{Config.Indent}ON {self.Schema}.{self.Table}"
        r += Config.NL
        r += f"{Config.Indent}AS {self.Type}"
        r += Config.NL
        r += f"{Config.Indent}FOR {self.Command}"
        r += Config.NL
        r += f"{Config.Indent}TO {self.Roles}"
        r += Config.NL

        if len(self.ExprUsing) > 0:
            r += f"{Config.Indent}USING({self.ExprUsing});"
        else:
            r += f"{Config.Indent}WITH CHECK({self.ExporWithCheck});"

        return r.strip()

    def Diff(self, another):
        result = []

        if (
            self.Command != another.Command or
            self.Type != another.Type or
            self.Roles != another.Roles or
            self.ExprUsing != another.ExprUsing or
            self.ExporWithCheck != another.ExporWithCheck
        ):
            result.append(another.DDL_Drop())
            result.append(self.DDL_Create())

        return result



#!/usr/bin/python
# -*- coding: utf-8 -*-

import os
import json
import logging
from timeit import default_timer

# Default lines separator
SEP = chr(10)

def CheckPath(path):
    """
        Create folder structure if not exists
        @param path: Path to folder
    """
    result = []
    for p in path.split("/"):
        result.append(p)
        p = "/".join(result)

        if not os.path.exists(p):
            os.mkdir(p)

def ReadConfig(path):
    assert os.path.exists(path), \
        "Config file {0} is not exists".format(path)

    with open(path, "r", encoding="utf8") as rf:
        data = json.loads(rf.read())

        assert data is not None, \
            "Config file is empty"

        assert isinstance(data, dict), \
            "Config file must contains dict"

        assert len(data or {}) > 0, \
            "Config file is empty"

        return data

def ParseACL(acl, owner):
    result = []
    roles  = []

    for acl in (acl or []):
        spl = acl.split("=")
        if len(spl) != 2:
            continue

        role_name = spl[0].strip()
        if role_name == "":
            role_name = "public"

        permissions = []

        grants = spl[1].replace("/%s" % (owner), "").strip()
        if grants.strip() == "":
            permissions.append("ALL")
        elif grants == "U":
            permissions.append("USAGE")
        elif grants == "UC":
            permissions.append("ALL")
        elif grants == "arwdDxt":
            permissions.append("ALL")
        elif grants == "rwU":
            permissions.append("ALL")
        elif grants == "X":
            permissions.append("EXECUTE")
        else:
            if grants.find("r") >= 0:
                permissions.append("SELECT")
            if grants.find("a") >= 0:
                permissions.append("INSERT")
            if grants.find("w") >= 0:
                permissions.append("UPDATE")
            if grants.find("d") >= 0:
                permissions.append("DELETE")
            if grants.find("D") >= 0:
                permissions.append("TRUNCATE")
            if grants.find("x") >= 0:
                permissions.append("REFERENCES")
            if grants.find("t") >= 0:
                permissions.append("TRIGGER")

        result.append({
            "is_grant" : True,
            "perm"     : permissions,
            "role"     : role_name
        })

        roles.append(role_name)

    if "public" not in roles:
        result.append({
            "is_grant" : False,
            "perm"     : ["ALL"],
            "role"     : "public"
        })

    if owner not in roles:
        result.append({
            "is_grant" : True,
            "perm"     : ["ALL"],
            "role"     : owner
        })

    result.sort(key=lambda x: (x.get('is_grant') is False, x.get('role')))

    return result

def ParseOptions(opt, exclude_extra=[]):
    exclude_names = [""] + exclude_extra

    eq_idx = opt.find("=")
    if eq_idx < 0 or eq_idx+1 == len(opt):
        return None

    name = (opt[:eq_idx] or "").strip()
    if name.lower() in exclude_names:
        return None

    value = (opt[eq_idx+1:] or "").strip()
    if len(value) <= 0:
        return None

    return f"{name} '{value}'"

def CalcDuration(fnc):
    def wrapper(*args):
        time_start = default_timer()
        result = fnc(*args)
        logging.info("{0} duration = {1} sec".format(fnc.__name__, round(default_timer() - time_start, 2)))
        return result
    return wrapper

def SetupLogging(config):
    logging.basicConfig(
        level=logging.getLevelName(config.get("log_level") or "INFO"),
        format='%(asctime)s [%(levelname)s] %(message)s',
        handlers=[
            logging.StreamHandler(),
            #handlers.TimedRotatingFileHandler("log.log", backupCount=7, when='D', interval=1)
        ]
    )
